//SugarMemo Database Module

#include <PalmOS.h>
#include "SugarMemo.h"
#include "SugarMemoRsc.h"
#include "SugarMemoDB.h"
#include "SugarMemoTools.h"

#define HeaderBufferVolumn	(UInt16)100

//*******************************************************************
//	
//	Globals
//
//*******************************************************************
DmOpenRef 	CurrentDB = NULL;
UInt16 		CurrentRecord = noRecord;

//*******************************************************************
//
//	Internal function declaration
//
//*******************************************************************

static 	UInt16	WordUnpackedSize(WordRecordType *record);
static 	void 	PackRecord(WordRecordType *record, MemPtr recordDBEntry);
static 	void 	UnpackRecord(WordPackedRecord *packed, WordRecordType *record);
static 	Int16 	WordDBComparePackedRecords(WordPackedRecord * packed1, WordPackedRecord * packed2,
	Int16 uselessinfo, SortRecordInfoPtr infoP1, SortRecordInfoPtr infoP2, MemHandle appinfoH);
static void TranslatePhonetic(char* field);

//*******************************************************************
//
//	App info manipulation. 	
//	1.Get AppInfo Pointer
//	2.Init AppInfo Pointer
//	3.GetRawWordCount(); 
//	4.SetRawWordCount();
//
//*******************************************************************

//return locked app info pointer
SugarAppInfoType*   GetAppInfoPtr(DmOpenRef dbP)
{
	UInt16     cardNo;
	LocalID    dbID;
	LocalID    appInfoID;

	if (DmOpenDatabaseInfo(dbP, &dbID, NULL, NULL, &cardNo, NULL))
		return NULL;
	if (DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL))
		return NULL;

	if (appInfoID == 0)
		return NULL;
	else
		return MemLocalIDToLockedPtr(appInfoID, cardNo);

}

Err	SugarDBAppInfoInit(DmOpenRef db){

	UInt16         		cardNo;
	LocalID        		dbID;
	LocalID        		appInfoID;
	MemHandle      		h;
	SugarAppInfoType* 	app;
	SugarAppInfoType* 	defaultP;
	UInt8          		i;
	UInt16				bits = 0;
	DateType			zeroDate;
	UInt16				version = 0;

	app = GetAppInfoPtr(db);
	if (app != NULL) {
		version = app -> version;
		MemPtrUnlock(app);
	}

	//	if version check OK, return
	if (version == CurrentVersion) return 0;

	//	clear old app info
	if (DmOpenDatabaseInfo(db, &dbID, NULL, NULL, &cardNo, NULL)) return dmErrInvalidParam;
	DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

	// Make space for new one
	h = DmNewHandle(db, sizeof(SugarAppInfoType));
	if (!h) return dmErrMemError;

	appInfoID = MemHandleToLocalID(h);
		
	if (DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL))
	{
		MemHandleFree(h);
		return 1;
	}

	app = MemHandleLock(h);


	// Allocate & Clear the app info
	defaultP = MemPtrNew(sizeof(SugarAppInfoType));
	if (defaultP == NULL)
	{
		ErrDisplay("Unable to init App info");
		return 1;
	}

	MemSet(defaultP, sizeof(SugarAppInfoType), 0);


/*
typedef struct {
	UInt16				renamedCategories;	// bitfield of categories with a different name
	char 				categoryLabels[dmRecNumCategories][dmCategoryLength];
	UInt8 				categoryUniqIDs[dmRecNumCategories];
	UInt8				lastUniqID;			// Uniq IDs generated by the device are between
											// 0 - 127.  Those from the PC are 128 - 255.
	UInt8				padding;			// from the compiler word aligning things
	
	DateType			lastUpdated;
	WordFieldFlags		answerPageFields;
	WordFieldFlags		questionPageFields;
	UInt16 				rawWordCount;
	
	UInt16				version;
	
	FontID				learnFieldFont[WordFieldsCount];
	FontID				learnFieldFont2[WordFieldsCount];
	
	TextAlignment		learnFieldAlignment[WordFieldsCount];	
	
	ReviewMode			reviewMode;
	UInt8				lefty;
	
}	SugarAppInfoType; */

	// Init the categories
	for (i = 0; i < dmRecNumCategories; i++)
	{
		defaultP->categoryUniqIDs[i] = i;
	}
		
	defaultP->lastUniqID = dmRecNumCategories - 1;

	//	Set lastUpdated
	DateSecondsToDate(0, &zeroDate);
	defaultP->lastUpdated = zeroDate;

	// 	Set question page field
	bits = 0;
	SetBitMacro(bits, WordFieldQuestion);
	SetBitMacro(bits, WordFieldPhonetic);
	defaultP -> questionPageFields.allBits = bits;

	// 	Set answer page field
	bits = 0;
	SetBitMacro(bits, WordFieldPhonetic);
	SetBitMacro(bits, WordFieldNote3);
	defaultP -> answerPageFields.allBits = bits;

	// 	Set raw word count to 0
	defaultP -> rawWordCount = 0;

	//	Set version
	defaultP -> version = CurrentVersion;

		
	//	Set LearnFieldFont
	defaultP -> learnFieldFont[WordFieldQuestion] = OS5_16_Font;
	defaultP -> learnFieldFont[WordFieldPhonetic] = PhoneticLargeFont;
	defaultP -> learnFieldFont[WordFieldAnswer] = OS5_16_Font;
	defaultP -> learnFieldFont[WordFieldWordNet] = stdFont;
	defaultP -> learnFieldFont[WordFieldMerriamWebster] = stdFont;
	defaultP -> learnFieldFont[WordFieldDict1] = stdFont;
	defaultP -> learnFieldFont[WordFieldDict2] = stdFont;
	defaultP -> learnFieldFont[WordFieldDict3] = stdFont;
	defaultP -> learnFieldFont[WordFieldNote1] = stdFont;
	defaultP -> learnFieldFont[WordFieldNote2] = stdFont;
	defaultP -> learnFieldFont[WordFieldNote3] = boldFont;

	//	Set LearnFieldFont2
	defaultP -> learnFieldFont2[WordFieldQuestion] = OS5_16_Font;
	defaultP -> learnFieldFont2[WordFieldPhonetic] = PhoneticLargeFont;
	defaultP -> learnFieldFont2[WordFieldAnswer] = OS5_16_Font;
	defaultP -> learnFieldFont2[WordFieldWordNet] = stdFont;
	defaultP -> learnFieldFont2[WordFieldMerriamWebster] = stdFont;
	defaultP -> learnFieldFont2[WordFieldDict1] = stdFont;
	defaultP -> learnFieldFont2[WordFieldDict2] = stdFont;
	defaultP -> learnFieldFont2[WordFieldDict3] = stdFont;
	defaultP -> learnFieldFont2[WordFieldNote1] = stdFont;
	defaultP -> learnFieldFont2[WordFieldNote2] = stdFont;
	defaultP -> learnFieldFont2[WordFieldNote3] = boldFont;

	//	Set LearnField text alignment
	defaultP -> learnFieldAlignment[WordFieldQuestion] = CenterAlignment;
	defaultP -> learnFieldAlignment[WordFieldPhonetic] = CenterAlignment;
	defaultP -> learnFieldAlignment[WordFieldAnswer] = CenterAlignment;
	defaultP -> learnFieldAlignment[WordFieldWordNet] = CenterAlignment;
	defaultP -> learnFieldAlignment[WordFieldMerriamWebster] = CenterAlignment;
	defaultP -> learnFieldAlignment[WordFieldDict1] = CenterAlignment;
	defaultP -> learnFieldAlignment[WordFieldDict2] = CenterAlignment;
	defaultP -> learnFieldAlignment[WordFieldDict3] = CenterAlignment;
	defaultP -> learnFieldAlignment[WordFieldNote1] = CenterAlignment;
	defaultP -> learnFieldAlignment[WordFieldNote2] = CenterAlignment;
	defaultP -> learnFieldAlignment[WordFieldNote3] = LeftAlignment;		

	//	Set review mode
	bits = 0;
	SetBitMacro(bits, TranslationRecollection);
	SetBitMacro(bits, WordRecollection);
	SetBitMacro(bits, TranslationTest);
	SetBitMacro(bits, WordTest);
	defaultP -> reviewMode.allBits = (UInt8)bits;

	//	Set lefty
	defaultP -> lefty = (UInt8)0;
	
	// copy in the defaults and free the default app info
	DmWrite(app, 0, defaultP,  sizeof(SugarAppInfoType));
	MemPtrFree(defaultP);

	CategoryInitialize((AppInfoPtr)app, CategoryNamesAppInfoStr);

	// Unlock
	MemPtrUnlock(app);

	return 0;
}

//initialize database appinfo 
/*
Err   SugarDBAppInfoInit(DmOpenRef dbP)
{	
	
	UInt16         	cardNo;
	LocalID        	dbID;
	LocalID        	appInfoID;
	MemHandle      	h;
	SugarAppInfoType* 	appInfoP;
	SugarAppInfoType* 	defaultInfoP;
	UInt8          	i, j;
	WordFieldFlags 	flag, newAnswerPageFlag;
	DateType 		today;
	
	Int16			oldVersion = -1;
	Int16			currentVersion = 2;
	
	UInt16 			bits = 0;
	
	MemorizingStatus memoStatus = {2.50, //e-factor
		40, //difficulty 
		0, //total lapse
		0, //total recall
		0, //recent lapse
		0, //recent recall
		0, //step
		0, //last test
		0, //current test
		0,	//next test
		0,	//first learn
		0}; //create

	SetBitMacro(bits, TranslationRecollection);
	SetBitMacro(bits, WordRecollection);
	SetBitMacro(bits, TranslationTest);
	SetBitMacro(bits, WordTest);
	
	newAnswerPageFlag.allBits = 0;
	SetBitMacro(newAnswerPageFlag.allBits, WordFieldPhonetic);
	SetBitMacro(newAnswerPageFlag.allBits, WordFieldNote3);


	DateDaysToDate(TimGetSeconds(), &today);
	
	appInfoP = GetAppInfoPtr(dbP);

	// If there isn't an app info make space for one
	if (appInfoP == NULL)
	{
	}
	
	else if (appInfoP -> version == currentVersion){
		MemPtrUnlock(appInfoP);
		return 0;
	}
	
	else {
		oldVersion = appInfoP -> version;
	}
	
	// Allocate & Clear the app info
	defaultInfoP = MemPtrNew(sizeof(SugarAppInfoType));
	if (defaultInfoP == NULL)
	{
		ErrDisplay("Unable to init Word DB");
		return 1;
	}
	
	// Flush memory
	MemSet(defaultInfoP, sizeof(SugarAppInfoType), 0);
	
	// Init the categories
	switch(oldVersion){
	
		case  -1:
			for (i = 0; i < dmRecNumCategories; i++)
				defaultInfoP->categoryUniqIDs[i] = i;
	
			defaultInfoP->lastUniqID = dmRecNumCategories - 1;
	
			//Set lastUpdated
			defaultInfoP -> lastUpdated = today;
	
			// Set question and answer page field
			flag.allBits = 0;
			SetBitMacro(flag.allBits, WordFieldQuestion);
			SetBitMacro(flag.allBits, WordFieldPhonetic);
			defaultInfoP -> questionPageFields.allBits = flag.allBits;
	
			//flag.allBits = 0;
			//SetBitMacro(flag.allBits, WordFieldAnswer);
			//SetBitMacro(flag.allBits, WordFieldNote3);
			defaultInfoP -> answerPageFields.allBits = newAnswerPageFlag.allBits;

			// Set raw word count to 0
			defaultInfoP -> rawWordCount = 0;
			defaultInfoP -> reviewMode.allBits = (UInt8)bits;
			break;
			
		case 0:
		case 1:
			defaultInfoP -> renamedCategories = appInfoP -> renamedCategories;	
			
			for (i = 0; i < dmRecNumCategories; i ++){
				defaultInfoP->categoryUniqIDs[i] = appInfoP -> categoryUniqIDs[i];

				for (j = 0; j < dmCategoryLength; j++){
					defaultInfoP -> categoryLabels[i][j] =	appInfoP -> categoryLabels[i][j];
				}
			}		
	
			defaultInfoP -> lastUniqID = appInfoP -> lastUniqID;
	
			//Set lastUpdated
			defaultInfoP -> lastUpdated = appInfoP -> lastUpdated;
	
			// Set question and answer page field
			defaultInfoP -> questionPageFields.allBits = 
				appInfoP -> questionPageFields.allBits;
	
			defaultInfoP -> answerPageFields.allBits = newAnswerPageFlag.allBits;

			// Set raw word count to 0
			defaultInfoP -> rawWordCount = appInfoP -> rawWordCount;
			defaultInfoP -> reviewMode.allBits = (UInt8)bits;
			break;
	}
		
	//init version
	defaultInfoP -> version = currentVersion;
		
	//init LearnFieldFont
	defaultInfoP -> learnFieldFont[WordFieldQuestion] = OS5_16_Font;
	defaultInfoP -> learnFieldFont[WordFieldPhonetic] = PhoneticLargeFont;
	defaultInfoP -> learnFieldFont[WordFieldAnswer] = OS5_16_Font;
	defaultInfoP -> learnFieldFont[WordFieldWordNet] = stdFont;
	defaultInfoP -> learnFieldFont[WordFieldMerriamWebster] = stdFont;
	defaultInfoP -> learnFieldFont[WordFieldDict1] = stdFont;
	defaultInfoP -> learnFieldFont[WordFieldDict2] = stdFont;
	defaultInfoP -> learnFieldFont[WordFieldDict3] = stdFont;
	defaultInfoP -> learnFieldFont[WordFieldNote1] = stdFont;
	defaultInfoP -> learnFieldFont[WordFieldNote2] = stdFont;
	defaultInfoP -> learnFieldFont[WordFieldNote3] = boldFont;

	//init LearnField text alignment
	defaultInfoP -> learnFieldAlignment[WordFieldQuestion] = CenterAlignment;
	defaultInfoP -> learnFieldAlignment[WordFieldPhonetic] = CenterAlignment;
	defaultInfoP -> learnFieldAlignment[WordFieldAnswer] = CenterAlignment;
	defaultInfoP -> learnFieldAlignment[WordFieldWordNet] = CenterAlignment;
	defaultInfoP -> learnFieldAlignment[WordFieldMerriamWebster] = CenterAlignment;
	defaultInfoP -> learnFieldAlignment[WordFieldDict1] = CenterAlignment;
	defaultInfoP -> learnFieldAlignment[WordFieldDict2] = CenterAlignment;
	defaultInfoP -> learnFieldAlignment[WordFieldDict3] = CenterAlignment;
	defaultInfoP -> learnFieldAlignment[WordFieldNote1] = CenterAlignment;
	defaultInfoP -> learnFieldAlignment[WordFieldNote2] = CenterAlignment;
	defaultInfoP -> learnFieldAlignment[WordFieldNote3] = LeftAlignment;		

	//Init default new record
	SetBitMacro(defaultInfoP -> defaultNewWordRecord.header.allBits, RawQueue);
	
	defaultInfoP -> defaultNewWordRecord.ltStatus.learningStatus = 0;
	defaultInfoP -> defaultNewWordRecord.memoStatus = memoStatus;
	defaultInfoP -> defaultNewWordRecord.memoStatus.create = today;

	
	if (appInfoP != NULL) {
		if (DmOpenDatabaseInfo(dbP, &dbID, NULL, NULL, &cardNo, NULL))
			return dmErrInvalidParam;
		if (DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL))
			return dmErrInvalidParam;		
		
		//delete original appInfo block
		if (DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL))
		{
			MemPtrUnlock(appInfoP);
			return 1;
		}
		
		//	free memory	
		MemPtrFree(appInfoP);
	}
	
	if (DmOpenDatabaseInfo(dbP, &dbID, NULL, NULL, &cardNo, NULL))
		return dmErrInvalidParam;
	if (DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL))
		return dmErrInvalidParam;
	
	//allocate new space
	h = DmNewHandle(dbP, sizeof(SugarAppInfoType));
	if (!h) return dmErrMemError;

	//set new app info block
	appInfoID = MemHandleToLocalID(h);
	if (DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &appInfoID, NULL, NULL, NULL))
	{
		MemHandleFree(h);
		return 1;
	}

	appInfoP = MemHandleLock(h);

	// copy in the defaults and free the default app info
	DmWrite(appInfoP, 0, defaultInfoP,  sizeof(SugarAppInfoType));
	MemPtrFree(defaultInfoP);

	// Try to use localized app info block strings.
	if (oldVersion == -1) ;

	// Unlock
	MemPtrUnlock(appInfoP);

	return 0;
}*/

UInt16 GetRawWordCount(){
	SugarAppInfoType* app;
	UInt16 number;
	
	app = GetAppInfoPtr(CurrentDB);
	number = app -> rawWordCount;
	MemPtrUnlock(app);
	
	return number;
}

void SetRawWordCount(UInt16 newWordCount){
	SugarAppInfoType* app;
	SugarAppInfoType* appHelper = 0;
	
	app = GetAppInfoPtr(CurrentDB);
	
	DmWrite(app, (UInt32)&appHelper -> rawWordCount, &newWordCount, sizeof(newWordCount));
	
	MemPtrUnlock(app);
}

void IncrementRawWordCount(){
	SugarAppInfoType* app;
	SugarAppInfoType* appHelper = 0;
	UInt16 count;
	
	app = GetAppInfoPtr(CurrentDB);
	count = app -> rawWordCount;
	count++;
	
	DmWrite(app, (UInt32)&appHelper -> rawWordCount, &count, sizeof(count));
	
	MemPtrUnlock(app);
}


void DecrementRawWordCount(){
	SugarAppInfoType* app;
	SugarAppInfoType* appHelper = 0;
	UInt16 count;
	
	app = GetAppInfoPtr(CurrentDB);
	count = app -> rawWordCount;
	if (count > 0) count--;
	
	DmWrite(app, (UInt32)&appHelper -> rawWordCount, &count, sizeof(count));
	
	MemPtrUnlock(app);
}
//*******************************************************************
//
//	Database operation
//	1.CreateWordDB(char* name);
//	2.RenameWordDB(char* name);
//	3.Boolean DeleteWordDB(UInt16 cardNo, LocalID dbID);
//
//*******************************************************************

//	create database
Boolean CreateWordDB(char* name){
	
	LocalID dbID;
	DmOpenRef db;
	UInt16 cardNo = 0; //internal mem
	UInt16 i;
	
	Err err = DmCreateDatabase(cardNo, name, appFileCreator, appVocabularyDB, false); //not resource db
	
	//popup error info
	if (err){ 
		ErrAlert(err);
		return false;
	}
	
	dbID = DmFindDatabase(cardNo, name);
	db = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);
	
	//initialize app info
	err = SugarDBAppInfoInit(db);
	
	if (err) {
		ErrAlert(err);
		DmCloseDatabase(db);
		DeleteWordDB(cardNo, dbID);
		return false;
	}
	
	for (i = TestingRecordIndex; i<HeaderIndexCounts; i++)
		CreateHeaderRecord(db, i);
	
	DmCloseDatabase(db);
	return true;
}

// 	rename database
Boolean RenameWordDB(char* originalName){
	return false;
}

//	delete databaase
Boolean DeleteWordDB(UInt16 cardNo, LocalID dbID){
	Err error;
	DmOpenRef db;
	UInt16 num, j;
	UInt32 progress, newProgress;
	Char message[20] = "Deleting....";
	
	db = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);
	num = DmNumRecords(db);
	
	progress = 0;
	ProgressPie(progress, message);
		
	for (j = 0; j < num; j++){
		DmDeleteRecord(db, num - 1 - j);
		
		newProgress = (j + 1) * ProgressFull / num;
		
		if (newProgress != progress){	
			progress = newProgress;
			ProgressPie(progress, message);
		}
	}
	
	DmCloseDatabase(db);
	
	error = DmDeleteDatabase(cardNo, dbID);	
	if (error) return false;
		
	return true;
}
//*******************************************************************
//
//	header record operation and maintenance
//
//	1.CreateHeaderRecord(DmOpenRef db, HeaderIndex headerIndex);
//	2.ClearHeaderRecord(DmOpenRef db, HeaderIndex headerIndex);
//	3.GetHeaderItemNumber(DmOpenRef db, HeaderIndex headerIndex);
//	4.GetHeaderItem(DmOpenRef db, HeaderIndex headerIndex, HeaderItem* ItemP, UInt16 index);
//	5.AddHeaderItem(DmOpenRef db, HeaderIndex headerIndex, HeaderItem* ItemP);
//	6.AddHeaderItemByDBIndex(DmOpenRef db, HeaderIndex headerIndex, UInt16 dbIndex);
//
//	7.RemoveHeaderItem(DmOpenRef db, HeaderIndex headerIndex, UInt16 index);
//	8.RemoveHeaderItemByID(DmOpenRef db, HeaderIndex headerIndex, UInt32 uniqueID);
//	9.RemoveHeaderItemByDBIndex(DmOpenRef db, HeaderIndex headerIndex, UInt16 dbIndex);
//	10.MoveHeaderItem(DmOpenRef db, HeaderIndex headerIndex, UInt16 index);
//
//	11.RebuildLearningQueue(DmOpenRef db);
//	12.UpdateTestingQueue();
//
//*******************************************************************


//This function create header record at certain index, so be sure to 
//create all header records in sequence when creating database, 
//wrong order may cause potential chaos
Boolean CreateHeaderRecord(DmOpenRef db, HeaderIndex headerIndex){

	MemHandle recordH;
	MemPtr recordP;
	HeaderRecord hRecord;	
	Err error;
	
	//prepare header flag
	hRecord.header.allBits = 0;
	SetBitMacro(hRecord.header.allBits, HeaderFlag);
	switch(headerIndex){
		case LearningRecordIndex:
			SetBitMacro(hRecord.header.allBits, LearningQFlag);
			break;
		case TestingRecordIndex:
			SetBitMacro(hRecord.header.allBits, TestingQFlag);
			break;
		case LearningChoiceBuffer:
			SetBitMacro(hRecord.header.allBits, LearningBFlag);
			break;
		case TestingChoiceBuffer:
			SetBitMacro(hRecord.header.allBits, TestingBFlag);
			break;
	}
	
	//set item number to 0
	hRecord.itemNumbers = 0;
	
	//allocate memory for an empty header record
	recordH = DmNewHandle(db, sizeof(HeaderRecord)); 
	
	if (recordH == NULL) return false;
	
	recordP = MemHandleLock(recordH); //lock once
	DmWrite(recordP, 0, &hRecord, sizeof(hRecord));
	MemPtrUnlock(recordP); //unlock once
		
	error = DmAttachRecord(db, &headerIndex, recordH, NULL);
	
	if (error) {
		MemHandleFree(recordH);	//release the memory resources if failed
		return false;
	}
	
	return true;
}

//clear header record, used for rebuild header record, 
//this function can be combined with create header record.
Boolean ClearHeaderRecord(DmOpenRef db, HeaderIndex headerIndex){

	MemHandle recordH, oldH;
	MemPtr recordP;
	HeaderRecord hRecord;	
	Err error;
	
	//prepare header flag
	hRecord.header.allBits = 0;
	SetBitMacro(hRecord.header.allBits, HeaderFlag);
	switch(headerIndex){
		case LearningRecordIndex:
			SetBitMacro(hRecord.header.allBits, LearningQFlag);
			break;
		case TestingRecordIndex:
			SetBitMacro(hRecord.header.allBits, TestingQFlag);
			break;
	}
	
	//set item number to 0
	hRecord.itemNumbers = 0;
	
	//allocate memory for an empty header record
	recordH = DmNewHandle(db, sizeof(HeaderRecord)); 
	
	if (recordH == NULL) return false;
	
	recordP = MemHandleLock(recordH); //lock once
	DmWrite(recordP, 0, &hRecord, sizeof(hRecord));
	MemPtrUnlock(recordP); //unlock once
		
	error = DmAttachRecord(db, &headerIndex, recordH, &oldH);
	
	if (error) {
		MemHandleFree(recordH);	//release the memory resources if failed
		return false;
	}
	
	MemHandleFree(oldH);
	
	return true;
}


//Get Item numbers in learning or testing queue
UInt16 GetHeaderItemNumber(DmOpenRef db, HeaderIndex headerIndex){
	
	UInt16 number;
	MemHandle recordH;
	HeaderRecord* recordP;
	
	recordH = DmQueryRecord(db, headerIndex); //error not handled 
	
	recordP = (HeaderRecord*)MemHandleLock(recordH);
	number = recordP -> itemNumbers;
	MemPtrUnlock(recordP);
	
	return number;
}

//Get learning or testing item
Boolean GetHeaderItem(DmOpenRef db, HeaderIndex headerIndex, HeaderItem* ItemP, UInt16 index){

	MemHandle recordH;
	HeaderRecordHelper* recordP;
	HeaderItem* p;
	
	recordH = DmQueryRecord(db, headerIndex);
	if (recordH == NULL) return false;
	
	recordP = (HeaderRecordHelper*)MemHandleLock(recordH);
	
	if (index > recordP -> itemNumbers - 1) return false;
	
	//point to the first item
	p = &recordP->firstItem;
	p += index;
	
	*ItemP = *p;
	
	MemPtrUnlock(recordP);
	
	return true;
}

//Add header item to tail
Boolean AddHeaderItem(DmOpenRef db, HeaderIndex headerIndex, HeaderItem* ItemP){

	MemHandle recordH;
	MemPtr recordP;
	UInt16 number, size, newSize;
	
	number = GetHeaderItemNumber(db, headerIndex);
	size = sizeof(HeaderRecord) + number * sizeof(HeaderItem);
	newSize = size + sizeof(HeaderItem);
	
	recordH = DmResizeRecord(db, headerIndex, newSize);
	if (recordH == NULL) return false;
	
	recordP = MemHandleLock(recordH);
	
	DmWrite(recordP, size, ItemP, sizeof(*ItemP));
	
	number++;
	DmWrite(recordP, sizeof(Header), &number, sizeof(number));
	
	MemHandleUnlock(recordH);
	
	return true;
}
	
	
	
	/*old function, using DmAttach method

	MemHandle recordH, oldH;
	MemPtr srcP, destP;
	HeaderRecordHelper* p = 0;
	UInt16 number, size, newSize, index;
	Err error;

	//calculate newsize
	number = GetHeaderItemNumber(db, headerIndex);
	size = sizeof(HeaderRecord) + number * sizeof(HeaderItem);
	newSize = size + sizeof(HeaderItem);
	
	//get original learning record
	srcP = MemHandleLock(DmQueryRecord(db, headerIndex)); //source locked
	
	//allocate new record			
	recordH = DmNewHandle(db, newSize);
	if (recordH == NULL) return false;
	
	destP = MemHandleLock(recordH);	//dest locked
	
	//copy original data
	DmWrite(destP, 0, srcP, size);

	//write new item
	DmWrite(destP, ((UInt32)&p->firstItem) + number * sizeof(HeaderItem), ItemP, sizeof(HeaderItem));
	
	//write new number
	number++;
	DmWrite(destP, (UInt32)&p->itemNumbers, &number, sizeof(number));
		
	//unlock src pointer
	MemPtrUnlock(srcP); //src unlocked
	
	//attach new record
	index = headerIndex;
	error = DmAttachRecord(db, &index, recordH, &oldH);
	
	//unlock dest pointer
	MemPtrUnlock(destP);
	
	if (error) {
		MemHandleFree(recordH);
		return false;
	}
	//release orphan data
	MemHandleFree(oldH);
	
	return true;
}*/



//	add header item by record index
Boolean AddHeaderItemByDBIndex(DmOpenRef db, HeaderIndex headerIndex, UInt16 dbIndex){
	
	HeaderItem item;
	
	DmRecordInfo (db, dbIndex, NULL, &item, NULL);
	return AddHeaderItem(db, headerIndex, &item);
	
}

Boolean ShuffleHeaderItem(DmOpenRef db, HeaderIndex headerIndex){
	
	MemHandle recordH;
	HeaderRecordHelper* recordP;
	HeaderItem item, *itemp1, *itemp2;
	UInt16 number, j1, j2, i, sizeofHeaderRecord = sizeof(HeaderRecord), sizeofHeaderItem = sizeof(HeaderItem);
	Err error;
	
	number = GetHeaderItemNumber(db, headerIndex);
	
	if (number == 0) return true;
	
	recordH = DmQueryRecord(db, headerIndex);
	if (recordH == NULL) return false;
	
	recordP = MemHandleLock(recordH);
	
	i = number;
	while (i > 0){
		j1 = SysRandom(0);
		j1 = (((UInt32)j1 * number) - 1) / sysRandomMax;
		j2 = SysRandom(0);
		j2 = (((UInt32)j2 * number) - 1) / sysRandomMax;
		
		if (j1 != j2){
		
			itemp1 = &recordP->firstItem;
			itemp1 += j1;
			
			itemp2 = &recordP->firstItem;
			itemp2 += j2;

			//save item 1
			item = *itemp1;			
			
			error = DmWriteCheck(recordP, sizeofHeaderRecord + j1 * sizeofHeaderItem, sizeofHeaderItem);
			if (error) {
				ErrAlert(error);
			}
			DmWrite(recordP, sizeofHeaderRecord + j1 * sizeofHeaderItem, itemp2, sizeofHeaderItem);
			
			error = DmWriteCheck(recordP, sizeofHeaderRecord + j2 * sizeofHeaderItem, sizeofHeaderItem);
			if (error) {
				ErrAlert(error);
			}
			DmWrite(recordP, sizeofHeaderRecord + j2 * sizeofHeaderItem, &item, sizeofHeaderItem);
		}
		
		i--;
	}
	
	MemHandleUnlock(recordH);
	
	return true;
}

/*
Boolean ShuffleHeaderItem(DmOpenRef db, HeaderIndex headerIndex){

	UInt16 number, size, i, random;
	MemHandle recordH, oldH;
	MemPtr srcP, destP;
	HeaderItem item;
	Err error;

	i = 0;
	number = GetHeaderItemNumber(db, headerIndex);
	size = sizeof(HeaderRecord) + number * sizeof(HeaderItem);
	
	recordH = DmNewHandle(db, size);
	if (recordH == NULL) return false;
	destP = MemHandleLock(recordH);

	//	copy header	
	srcP = MemHandleLock(DmQueryRecord(db, headerIndex));
	DmWrite(destP, 0, srcP, sizeof(HeaderRecord));
	MemPtrUnlock(srcP);
		
	while (number > 0){
		
		random = SysRandom(0);
		random = ((UInt32)random * (number - 1)) / sysRandomMax;
		
		GetHeaderItem(db, headerIndex, &item, random);
		
		DmWrite(destP, sizeof(HeaderRecord) + i * sizeof(HeaderItem), &item, sizeof(item));
		
		RemoveHeaderItem(db, headerIndex, random);
		
		i++;
		number--;
	}
	
	error = DmAttachRecord(db, &headerIndex, recordH, &oldH);
	
	MemPtrUnlock(destP);
	
	if (error){
	
		MemHandleFree(recordH);
		return false;
	
	}
	
	MemHandleFree(oldH);
	
	return true;
}*/


//remove the i th item in queue
Boolean RemoveHeaderItem(DmOpenRef db, HeaderIndex headerIndex, UInt16 i){

	MemHandle recordH, oldH;
	MemPtr destP;
	HeaderItem* itemP;
	HeaderRecordHelper* srcP;
	HeaderRecordHelper* p = 0;
	UInt16 numbers, size, newSize, offset;
	Err error;
	
	//calculate newsize
	numbers = GetHeaderItemNumber(db, headerIndex);
	size = sizeof(HeaderRecord) + numbers * sizeof(HeaderItem);
	
	newSize = size - sizeof(HeaderItem);
	
	//get original learning record
	srcP = (HeaderRecordHelper*)MemHandleLock(DmQueryRecord(db, headerIndex));	//src locked
	
	//allocate new record			
	recordH = DmNewHandle(db, newSize);
	if (recordH == NULL) return false;
	
	destP = MemHandleLock(recordH);	//dest locked
	
	//copy original data before item i
	DmWrite(destP, 0, srcP, sizeof(HeaderRecord) + i * sizeof(HeaderItem));

	//copy orginal data after item i, if any
	if (i != (numbers - 1)){
		itemP = &srcP->firstItem;
		itemP += i + 1;
		offset = (UInt32)&p->firstItem + i * sizeof(HeaderItem);
		DmWrite(destP, offset, itemP, (numbers - i - 1) * sizeof(HeaderItem));
	}
	//write new number
	numbers--;
	DmWrite(destP, (UInt32)&p->itemNumbers, &numbers, sizeof(numbers));
		
	//unlock src pointer
	MemPtrUnlock(srcP);
	
	//attach new record
	error = DmAttachRecord(db, &headerIndex, recordH, &oldH);
	
	//unlock dest pointer
	MemPtrUnlock(destP);
	
	if (error) {
		MemHandleFree(recordH);
		return false;
	}
	
	//release orphan data
	MemHandleFree(oldH);
	
	return true;
}

// 
Boolean FindHeaderItemByID(DmOpenRef db, HeaderIndex headerIndex, UInt32 uniqueID, UInt16* indexP){

	MemHandle recordH;
	HeaderRecordHelper* recordP;
	HeaderItem* p;
	UInt16 numbers, i = 0;
	Boolean found = false;
	
	recordH = DmQueryRecord(db, headerIndex);
	if (recordH == NULL) return false;
	
	recordP = (HeaderRecordHelper*)MemHandleLock(recordH);
	numbers = recordP -> itemNumbers;
	
	p = &recordP -> firstItem;
	while (i < numbers){
		if ((UInt32)(*p)== uniqueID) {
			found = true;
			break;
		}
		p++;
		i++;
	}
	
	MemPtrUnlock(recordP);
	
	if (found) {
		(*indexP) = i;
		return true;
	}
	
	return false;
}

Boolean FindHeaderItemByDBIndex(DmOpenRef db, HeaderIndex headerIndex, UInt16 dbIndex, UInt16* indexP){

	HeaderItem item;
	
	DmRecordInfo(db, dbIndex, NULL, &item, NULL);
	
	return FindHeaderItemByID(db, headerIndex, item, indexP);

}

//	remove header item according item content, i.e. unique ID
Boolean RemoveHeaderItemByID(DmOpenRef db, HeaderIndex headerIndex, UInt32 uniqueID){
	
	UInt16 i;
	
	if (FindHeaderItemByID(db, headerIndex, uniqueID, &i)) return RemoveHeaderItem(db, headerIndex, i);

	return false;
}

//remove header item by database index
Boolean RemoveHeaderItemByDBIndex(DmOpenRef db, HeaderIndex headerIndex, UInt16 dbIndex){

	HeaderItem item;
	
	DmRecordInfo(db, dbIndex, NULL, &item, NULL);
	
	return RemoveHeaderItemByID(db, headerIndex, item);
}

//remove the first one, insert it into certain position, the status changed in LTStatus
Boolean MoveHeaderItem(DmOpenRef db, HeaderIndex headerIndex, UInt16 index){

	MemHandle recordH;
	HeaderRecordHelper* recordP;
	HeaderItem* p;
	HeaderItem item0;
	UInt16 i;
	
	//check index range
	if (index > GetHeaderItemNumber(db, headerIndex) - 1)
		index = GetHeaderItemNumber(db, headerIndex) - 1;
	
	//save first item
	GetHeaderItem(db, headerIndex, &item0, 0);
	
	recordH = DmQueryRecord(db, headerIndex);
	if (recordH == NULL) return false;
	
	recordP = (HeaderRecordHelper*)MemHandleLock(recordH);
	
	//point to item 1, the second item
	p = &recordP -> firstItem;
	p++; //pointer to item 1
	
	//move 
	for (i = 0; i < index; i++){
		DmWrite(recordP, (sizeof(HeaderRecord) + i * sizeof(HeaderItem)), p, sizeof(HeaderItem));
		p++;
	}
	
	//restore first item
	DmWrite(recordP, (sizeof(HeaderRecord) + index * sizeof(HeaderItem)), &item0, sizeof(HeaderItem));
	
	MemPtrUnlock(recordP);
		
	return true;
}

//	rebuild testing queue
void RebuildTestingQueue(DmOpenRef db, Boolean forced){
	
	UInt16 j, num;
	UInt16 rawWordCount = 0;
	DateType today;
	
	WordPackedRecord* packed;
	MemHandle recordH;
	HeaderItem	item;
	
	SugarAppInfoType* app;
	SugarAppInfoType* appHelper = 0;
	
	Header flag;

	UInt32 progress = 0;
	UInt32 newProgress;
	Char title[dmDBNameLength + 20];
	Char dbName[dmDBNameLength + 1] = "Unknown";
	LocalID localID;
	Err error;
		
	
	DateSecondsToDate(TimGetSeconds(), &today);
	
	app = GetAppInfoPtr(db);
	if (app == NULL) {
		DoCustomDialog("", "޷AppInfo", false, 0, NULL, false);;		
		return;
	}
	
	if (!forced){
		if (DateToDays(app->lastUpdated) == DateToDays(today)) {
			MemPtrUnlock(app);
			return;
		}
	}		
	
	ClearHeaderRecord(db, TestingRecordIndex);
				
	//loop records
	num = DmNumRecordsInCategory(db, dmAllCategories);
	
	if (num > FirstWordRecord) {
		error = DmOpenDatabaseInfo(db, &localID, NULL, NULL, NULL, NULL);
		
		if (error) ErrAlert(error);
		else {
			
			error = DmDatabaseInfo(0, localID, dbName, NULL, NULL, NULL, NULL, NULL, 
				NULL, NULL, NULL, NULL, NULL);
		
			if (error) ErrAlert(error);
			else StrPrintF(title, "%s %s", "Updating", dbName);
			
		}
		ProgressPie(progress, title);
	}		
	
	for (j = FirstWordRecord; j < num; j++){
			
		recordH = DmQueryRecord(db, j);
		if (recordH == NULL) ErrAlert(DmGetLastErr());
		else {
		
			packed = MemHandleLock(recordH);
			flag.allBits = packed -> header.allBits;
			
			if (GetBitMacro(flag.allBits, RawQueue)) {
				MemHandleUnlock(recordH);
				rawWordCount ++;
			}
				
			else if (GetBitMacro(flag.allBits, TestingQueue)){
				MemHandleUnlock(recordH);
				DmRecordInfo(db, j, NULL, &item, NULL);
				AddHeaderItem(db, TestingRecordIndex, &item);
			}
			
			else if (GetBitMacro(flag.allBits, WaitingQueue)){	
				
				//	if set queue before unlock mem, free handle error !
				if (DateToDays(packed -> memoStatus.nextTest) <= DateToDays(today)){
					MemHandleUnlock(recordH);
					SetQueue(db, j, TestingQueue);
					//DmRecordInfo(db, j, NULL, &item, NULL);
					//AddHeaderItem(db, TestingRecordIndex, &item);
				}
				else MemHandleUnlock(recordH);
			}
			
			//MemHandleUnlock(recordH);
		}
		
		newProgress = (j - 1) * ProgressFull / (num - FirstWordRecord);
		if (newProgress != progress) {
			progress = newProgress;
			ProgressPie(progress, title);
		}
	}
	
	ShuffleHeaderItem(db, TestingRecordIndex);
		
	//updage sugar db app info
	DmWrite(app, (UInt32)&appHelper->lastUpdated, &today, sizeof(today));
	DmWrite(app, (UInt32)&appHelper->rawWordCount, &rawWordCount, sizeof(rawWordCount));
	MemPtrUnlock(app);		
		
}

//	rebuild learning queue
void RebuildLearningQueue(DmOpenRef db){

	MemHandle recordH;
	WordPackedRecord* packed;
	WordPackedRecord* helper = 0;
	LearningStatus ls = FirstLook;
	HeaderItem item;
	UInt16 j;
	UInt16 num;
	
	//clear Learing record
	ClearHeaderRecord(db, LearningRecordIndex);
		
	//loop records
	num = DmNumRecordsInCategory(db, dmAllCategories);
	for (j = FirstWordRecord; j < num; j++){
			
		recordH = DmQueryRecord(db, j);
		packed = MemHandleLock(recordH);
		
		if (GetBitMacro(packed->header.allBits, LearningQueue)){
			DmRecordInfo (db, j, NULL, &item, NULL);
			AddHeaderItem(db, LearningRecordIndex, &item);
			DmWrite(packed, (UInt32)&helper -> ltStatus.learningStatus, &ls, sizeof (LearningStatus));
		}
					
		MemHandleUnlock(recordH);
	}
}

void ResetDatabase(DmOpenRef db){
	
	SugarAppInfoType* app;
	SugarAppInfoType* helper = 0;
	WordPackedRecord packed;
	MemHandle recordH;
	MemPtr p;
	UInt16 j, num;
	UInt32 progress = 0;
	UInt32 newProgress;
	char progressTitle[20] = "Clearing....";
	
	MemSet(&packed, sizeof(packed), 0);
	
	SetBitMacro(packed.header.allBits, RawQueue);
	packed.memoStatus.eFactor = 2.5;
	packed.memoStatus.difficulty = 40;
	
	app = GetAppInfoPtr(db);
	num = DmNumRecordsInCategory(db, dmAllCategories);
	
	if (num > FirstWordRecord) ProgressPie(progress, progressTitle);
	
	for (j = FirstWordRecord; j < num; j++){
	
		recordH = DmQueryRecord(db, j);
		p = MemHandleLock(recordH);
	
		DmWrite(p, 0, &packed,  
			//header + LTStatus + MemorizingStatus - CreateDate (createDate not changed)
			sizeof(Header) + sizeof(LTStatus) + sizeof(MemorizingStatus) - sizeof(DateType));
			
		MemHandleUnlock(recordH);
		
		newProgress = (j - 1) * ProgressFull / (num - FirstWordRecord);
		if (newProgress != progress) {
			progress = newProgress;
			ProgressPie(progress, progressTitle);
		}
	}
	if (num > FirstWordRecord) num -= FirstWordRecord;
	else num = 0;
	
	DmWrite(app, (UInt32)&helper -> rawWordCount, &num, sizeof(num));
	
	MemPtrUnlock(app);
	
	ClearHeaderRecord(db, LearningRecordIndex);
	ClearHeaderRecord(db, TestingRecordIndex);
	ClearHeaderRecord(db, LearningChoiceBuffer);
	ClearHeaderRecord(db, TestingChoiceBuffer);
}

//*******************************************************************
//
//	Word record manipulation
//	1.WordUnpackedSize(WordRecordType *record);
//	2.PackRecord(WordRecordType *record, MemPtr recordDBEntry);
//	3.UnpackRecord(WordPackedRecord *packed, WordRecordType *record);
//	4.WordDBComparePackedRecords(WordPackedRecord * packed1, WordPackedRecord * packed2,
//		Int16 uselessinfo, SortRecordInfoPtr infoP1, SortRecordInfoPtr infoP2, MemHandle appinfoH);
//
//	5.NewWordRecord(DmOpenRef db, WordRecordType* record, UInt16* index);
//	6.GetWordRecord(DmOpenRef db, UInt16 index, WordRecordType* recordP, MemHandle *recordH);
//	7.WordDBDeleteRecord(Boolean archive);
//	8.ChangeWordRecord(DmOpenRef dbP, UInt16 *index, WordRecordType* r, WordFieldFlags changedFields);
//
//*******************************************************************


//	calculate the unpacked record size
static UInt16 WordUnpackedSize(WordRecordType *record){

	UInt16 size;
	int i;
	
	//initial size is the size of a packed record, minus a placeholder of first field
	size = sizeof(WordPackedRecord) - sizeof(char);
	
	for (i = 0; i < WordFieldsCount; i++){
		if (record->field[i] != NULL)
			size += StrLen(record->field[i])+1;
	}
	
	return size;
}

//	pack a record
static void PackRecord(WordRecordType *record, MemPtr recordDBEntry){
	UInt32 offset;
	Int16 index;
	UInt16 length;
	MemPtr p;
	WordFieldFlags flags;
	WordPackedRecord *packed = 0;		//set the pointer to zero! this pointer is used to calculate offset position,
													//NEVER try to use this pointer to access data, will cause fatal error!
	
	UInt16 answerOffset = 0;
	
	flags.allBits = 0;
	
	//write wordstatus first, noting that (UInt32)&packed->wordStatus convert the reference pointer to a UInt32 number,
	//since the packed address is originally set to zero, &packed->wordStatus point to the starting position of 
	//wordStatus, which is zero also at this time. Dangerous but efficient usage!
	 
	DmWrite(recordDBEntry, (UInt32)&packed->header, &record->header, sizeof(record->header));
	DmWrite(recordDBEntry, (UInt32)&packed->ltStatus, &record->ltStatus, sizeof(record->ltStatus));
	DmWrite(recordDBEntry, (UInt32)&packed->memoStatus, &record->memoStatus, sizeof(record->memoStatus));
	
	//pointer to first field placeholder
	offset = (UInt32)&packed->firstField;
	
	//write all fields into record and set the flag bit
	for (index = WordFieldQuestion; index < WordFieldsCount; index++){
		if (record->field[index]!=NULL){
		p = record->field[index];
		length = StrLen(p)+1;
		DmWrite(recordDBEntry, offset, p, length);
		offset += length;
		SetBitMacro(flags.allBits, index);
		}
	
	}	//end of loop
	
	//write flags
	DmWrite(recordDBEntry, (UInt32)&packed->wordDBRecordFlags, &flags.allBits, sizeof(flags.allBits));
	
	//set or clear field offset
	index = 0;
	
	if (record->field[WordFieldQuestion]!=NULL)
		index += StrLen(record->field[WordFieldQuestion])+1;
	if (record->field[WordFieldPhonetic]!=NULL)
		index += StrLen(record->field[WordFieldPhonetic])+1;
	if (record->field[WordFieldAnswer]!=NULL)
		answerOffset = index;
	
	DmWrite(recordDBEntry, (UInt32)&packed->answerOffset, &answerOffset, sizeof(answerOffset));

}	//end of routine PackRecord

//	unpack word record
static void UnpackRecord(WordPackedRecord *packed, WordRecordType *record){
	Int16 index;
	UInt16 flags;
	char* p;
	
	record->header = packed ->header;
	record->ltStatus = packed ->ltStatus;
	record->memoStatus = packed ->memoStatus;
	flags = packed -> wordDBRecordFlags.allBits;
	p = & packed -> firstField;
	
	for (index = 0; index < WordFieldsCount; index ++){
		if (GetBitMacro(flags, index) != 0){
			record->field[index] = p;
			p += StrLen(p) + 1;
		}
		else{
			record->field[index] = NULL;
		}
	}
}

//	callback funtion for compare records
static Int16 WordDBComparePackedRecords(WordPackedRecord* packed1, WordPackedRecord* packed2,
	Int16 uselessinfo, SortRecordInfoPtr infoP1, SortRecordInfoPtr infoP2, MemHandle appinfoH){
	
	//if both header remain their sequence
	if ((GetBitMacro(packed1->header.allBits, HeaderFlag))
			&& (GetBitMacro(packed2->header.allBits, HeaderFlag)))
		return 0;
	
	//otherwise the header record is less than the other one
	if (GetBitMacro(packed1->header.allBits, HeaderFlag))
		return -1;
	
	if (GetBitMacro(packed2->header.allBits, HeaderFlag))
		return 1;
	
	//both word, and both no question field
	if ((GetBitMacro(packed1->wordDBRecordFlags.allBits, WordFieldQuestion) == 0)
		&& (GetBitMacro(packed2->wordDBRecordFlags.allBits, WordFieldQuestion) == 0))
		return 0;
	
	//both word, the one without question field is less than the other one
	if (GetBitMacro(packed1->wordDBRecordFlags.allBits, WordFieldQuestion) == 0)
		return -1;
	
	if (GetBitMacro(packed2->wordDBRecordFlags.allBits, WordFieldQuestion) == 0)
		return 1;
	
	//both word and both have the question field, compare string
	return StrCaselessCompare((char*)(&packed1->firstField), (char*)(&packed2->firstField));
}

//	create new record
Err NewWordRecord(DmOpenRef db, WordRecordType* record, UInt16* index){
	
	MemHandle recordH;
	WordPackedRecord* packed;
	UInt16 newIndex;
	Err error;
	
	//allocate new record size
	recordH = DmNewHandle(db, WordUnpackedSize(record));
	
	if (recordH == NULL) return dmErrMemError;
	
	//copy the data from unpacked record to packed one
	packed = MemHandleLock(recordH); //locked once
	PackRecord(record, packed);
	
	//get the index of new record
	newIndex = DmFindSortPosition(db, packed, NULL, (DmComparF*)WordDBComparePackedRecords, NULL);
	MemPtrUnlock(packed); //unlocked
	
	//attach new record in place and return the new index
	error = DmAttachRecord(db, &newIndex, recordH, NULL);
	
	if (error)
		MemHandleFree(recordH);
	else
		*index = newIndex;
	
	return error;
}

//	Get Word Record, noting that recordH is locked here, should be unlock else where!
Err GetWordRecord(DmOpenRef db, UInt16 index, WordRecordType* recordP, MemHandle *recordH)
{
	WordPackedRecord *packed;

	*recordH = DmQueryRecord(db, index);
	packed = (WordPackedRecord *) MemHandleLock(*recordH);
	if (packed == NULL)
		return dmErrIndexOutOfRange;

	UnpackRecord(packed, recordP);

	return 0;
}

//	Change Word Record
Err ChangeWordRecord(DmOpenRef dbP, UInt16 *index, WordRecordType* r, WordFieldFlags changedFields)
{
	WordRecordType     src;
	MemHandle          srcH;
	Err                error;
	MemHandle          recordH = 0; ///???
	MemHandle          oldH;
	Int16              i;
	UInt32             changes = changedFields.allBits;
	Boolean            dontMove;
	UInt16             attributes;      // to contain the deleted flag

	WordPackedRecord*   cmpP;
	WordPackedRecord*   recordP;

	// We do not assume that r is completely valid so we get a valid
	// WordRecordType*...
	if ((error = GetWordRecord(dbP, *index, &src, &srcH)) != 0)
		return error;

	// and we apply the changes to it.
	src.header = r->header;         // copy the status
	src.ltStatus = r->ltStatus;
	src.memoStatus = r->memoStatus;
	
	for (i = WordFieldQuestion; i < WordFieldsCount; i++)
	{
		// If the flag is set point to the string else NULL
		if (GetBitMacro(changes, i) != 0)
		{
			src.field[i] = r->field[i];
			RemoveBitMacro(changes, i);
		}
		if (changes == 0)
			break;      // no more changes
	}

	
	// 1) and 2) (make a new chunk with the correct size)
	recordH = DmNewHandle(dbP, WordUnpackedSize(&src));
	if (recordH == NULL)
	{
		MemHandleUnlock(srcH);      // undo lock from GetWordRecord above
		return dmErrMemError;
	}
	recordP = MemHandleLock(recordH);


	// 3) Copy the data from the unpacked record to the packed one.
	PackRecord(&src, recordP);

	// The original record is copied and no longer needed.
	MemHandleUnlock(srcH);


	// 4) if the sort position changes...
	// Check if question field has changed
	if ((changedFields.allBits & BitAtPosition(WordFieldQuestion)) == 0)
		goto attachRecord;

	
	// Make sure *index-1 < *index < *index+1, if so it's in sorted
	// order.  Leave it there.
	//appInfoPtr = (SugarAppInfoType*) GetAppInfoPtr(dbP);
	//sortByCompany = appInfoPtr->misc.sortByCompany;
	//MemPtrUnlock(appInfoPtr);

	if (*index > HeaderIndexCounts)
	{
		// This record wasn't deleted and deleted records are at the end of the
		// database so the prior record may not be deleted!
		cmpP = MemHandleLock(DmQueryRecord(dbP, *index-1));
		dontMove = (WordDBComparePackedRecords (cmpP,  recordP, NULL, NULL, NULL, 0) < 0);
		MemPtrUnlock(cmpP);
	}
	else
		dontMove = true;


	if (*index+1 < DmNumRecords (dbP))
	{
		DmRecordInfo(dbP, *index+1, &attributes, NULL, NULL);
		if (attributes & dmRecAttrDelete)
			;      // don't move it after the deleted record!
		else {
			cmpP = MemHandleLock(DmQueryRecord(dbP, *index+1));
			dontMove = dontMove && (WordDBComparePackedRecords (recordP, cmpP, NULL, NULL, NULL, 0) == -1);
			MemPtrUnlock(cmpP);
		}
	}


	if (dontMove)
		goto attachRecord;



	// The record isn't in the right position.  Move it.
	i = DmFindSortPosition(dbP, recordP, NULL, (DmComparF*)WordDBComparePackedRecords, NULL);
	DmMoveRecord(dbP, *index, i);
	if (i > *index) i--;
	*index = i;                  // return new position


	// Attach the new record to the old index,  the preserves the
	// category and record id.
	
	
attachRecord:

	error = DmAttachRecord(dbP, index, recordH, &oldH);
	MemPtrUnlock(recordP);
	if (error) return error;

	MemHandleFree(oldH);
	
	return 0;
}

// delete word record
void WordDBDeleteRecord(Boolean archive){
	
	HeaderQueue queue = GetQueue(CurrentDB, CurrentRecord);
	
	if (queue == RawQueue) 		DecrementRawWordCount();	
	
	if (queue == LearningQueue) RemoveHeaderItemByDBIndex(CurrentDB, LearningRecordIndex, CurrentRecord);
	if (queue == TestingQueue)  RemoveHeaderItemByDBIndex(CurrentDB, TestingRecordIndex, CurrentRecord);
	RemoveHeaderItemByDBIndex(CurrentDB, TestingChoiceBuffer, CurrentRecord);	
	RemoveHeaderItemByDBIndex(CurrentDB, LearningChoiceBuffer, CurrentRecord);
	
	if (archive)
	{
		DmArchiveRecord (CurrentDB, CurrentRecord);
	}
	else
		DmRemoveRecord (CurrentDB, CurrentRecord);

	CurrentRecord = noRecord;
}

//*******************************************************************
//
//	high-level word record manipulation
//	1.GetQueue();
//	2.SetQueue(HeaderQueue queue);
//	3.GetLearningStatus();
//	4.SetLearningStatus(LearningStatus ls);
//	5.GetTestingStatus();
//	6.SetTestingStatus(TestingStatus ls);
//	7.GetMemorizingStatus();
//	8.SetmemorizingStatus(MemorizingStatus memoStatus);
//
//*******************************************************************


//	Get CurrentRecord Queue
HeaderQueue GetQueue(DmOpenRef db, UInt16 index){

	HeaderQueue queue;
	Header* header;
	MemHandle recordH;
	
	recordH = DmQueryRecord(db, index);
	if (recordH == NULL) return BadQueue;
	
	header = MemHandleLock(recordH);

	if (GetBitMacro(header->allBits, RawQueue))	queue = RawQueue;
	else if (GetBitMacro(header->allBits, LearningQueue)) queue = LearningQueue;
	else if (GetBitMacro(header->allBits, TestingQueue)) queue = TestingQueue;
	else if (GetBitMacro(header->allBits, WaitingQueue)) queue = WaitingQueue;

	else if (GetBitMacro(header->allBits, HeaderFlag) && GetBitMacro(header->allBits, LearningQFlag)) 
		queue = LearningQFlag;
	else if (GetBitMacro(header->allBits, HeaderFlag) && GetBitMacro(header->allBits, TestingQFlag))
		queue = TestingQFlag;
	else if (GetBitMacro(header->allBits, HeaderFlag) && GetBitMacro(header->allBits, LearningBFlag)) 
		queue = LearningBFlag;
	else if (GetBitMacro(header->allBits, HeaderFlag) && GetBitMacro(header->allBits, TestingBFlag))
		queue = TestingBFlag;

	MemHandleUnlock(recordH);
	return queue;
}

//	Set CurrentRecord Queue
void	SetQueue(DmOpenRef db, UInt16 index, HeaderQueue queue){

	WordFieldFlags flag;
	WordRecordType record;
	MemHandle recordH;
	HeaderQueue oldQueue;
	
	oldQueue = GetQueue(db, index);
	
	if (oldQueue == queue) return;
	
	//	remove from old header Q 
	if (oldQueue == LearningQueue) 	RemoveHeaderItemByDBIndex(db, LearningRecordIndex, index);
	if (oldQueue == TestingQueue) 	RemoveHeaderItemByDBIndex(db, TestingRecordIndex, index);
		
	//	add to new header Q
	if (queue == LearningQueue) AddHeaderItemByDBIndex(db, LearningRecordIndex, index);
	if (queue == TestingQueue)	AddHeaderItemByDBIndex(db, TestingRecordIndex, index);
	
	//	update raw word count
	if (oldQueue == RawQueue) DecrementRawWordCount();
	if (queue == RawQueue) IncrementRawWordCount();
	
	flag.allBits = 0;
	
	GetWordRecord(db, index, &record, &recordH);
	record.header.allBits = BitAtPosition(queue);
	ChangeWordRecord(db, &index, &record, flag);

}

//	Get learning status of CurrentRecord
LearningStatus GetLearningStatus(DmOpenRef db, UInt16 index){
	WordRecordType record;
	MemHandle recordH;
	LearningStatus status;
	
	GetWordRecord(db, index, &record, &recordH); //lock recordH
	status = record.ltStatus.learningStatus;
	MemHandleUnlock(recordH);
	
	return status;
}

//	Set learning status of CurrentRecord
void SetLearningStatus(DmOpenRef db, UInt16 index, LearningStatus ls){
	WordFieldFlags flag;
	WordRecordType record;
	MemHandle recordH;
	
	GetWordRecord(db, index, &record, &recordH);
	record.ltStatus.learningStatus = ls;
	flag.allBits = 0;
	ChangeWordRecord(db, &index, &record, flag); 
}

//	Get testing status of CurrentRecord
TestingStatus GetTestingStatus(DmOpenRef db, UInt16 index){
	WordRecordType record;
	MemHandle recordH;
	TestingStatus status;
	
	GetWordRecord(db, index, &record, &recordH); //lock recordH
	status = record.ltStatus.testingStatus;
	MemHandleUnlock(recordH);
	
	return status;
}

//	Set testing status of CurrentRecord
void SetTestingStatus(DmOpenRef db, UInt16 index, TestingStatus ts){
	WordFieldFlags flag;
	WordRecordType record;
	MemHandle recordH;
	
	GetWordRecord(db, index, &record, &recordH);
	record.ltStatus.testingStatus = ts;
	flag.allBits = 0;
	ChangeWordRecord(db, &index, &record, flag); 
}

//	Get memorizing status of CurrentRecord
MemorizingStatus GetMemorizingStatus(DmOpenRef db, UInt16 index){
	WordRecordType record;
	MemHandle recordH;
	MemorizingStatus memoStatus;
	
	GetWordRecord(db, index, &record, &recordH);
	memoStatus = record.memoStatus;
	MemHandleUnlock(recordH);
	return memoStatus;
}

//	Set memorizing status of CurrentRecord
void SetMemorizingStatus(DmOpenRef db, UInt16 index, MemorizingStatus memoStatus){
	WordRecordType record;
	MemHandle recordH;
	WordFieldFlags flag;
	
	flag.allBits = 0;
	GetWordRecord(db, index, &record, &recordH);
	record.memoStatus = memoStatus;
	ChangeWordRecord(db, &index, &record, flag);
}

//*******************************************************************
//
//	Misc function relating to database manipulation
//	1.GetDBList();
//	2.WordRecordContainsData (WordRecordType* recordP);
//	3.InstantLookupPosition(DmOpenRef db, Char* p);
//	4.GetWordQueue(DmOpenRef db, UInt16 index);
//
//*******************************************************************

//	Buil
void InitHeaderBuffer(DmOpenRef db, HeaderIndex BufferIndex){
	
	MemHandle srcH, destH;
	HeaderRecord* srcP, *destP;
	MemPtr offsetP;
	UInt16 bufferNum, srcHeaderIndex, destHeaderIndex; // index; //, i, j;
	UInt32 newSize;
	//HeaderItem item;
	//Err error;

	//	validate	
	//bufferNum = GetHeaderItemNumber(db, BufferIndex);
	//for (i = 0; i < bufferNum; i++){
		
		//j = bufferNum -i - 1;
		//if (!GetHeaderItem(db, BufferIndex, &item, j)) continue;
		
		//error = DmFindRecordByID(db, item, &index);
		//if (error)	RemoveHeaderItem(db, BufferIndex, bufferNum - i - 1);

	//}

	bufferNum = GetHeaderItemNumber(db, BufferIndex);
		
	if (bufferNum > 0) return;
	
	if (BufferIndex == LearningChoiceBuffer){
		srcHeaderIndex = LearningRecordIndex;
		destHeaderIndex = LearningChoiceBuffer;	
	}
	
	else if (BufferIndex == TestingChoiceBuffer){
	
		if (GetHeaderItemNumber(db, LearningChoiceBuffer) > 0) 
			srcHeaderIndex = LearningChoiceBuffer;
		else if (GetHeaderItemNumber(db, TestingRecordIndex) > 0) 
			srcHeaderIndex = TestingRecordIndex;
			
		destHeaderIndex = TestingChoiceBuffer;
	}
	
	newSize = sizeof(HeaderRecord) + sizeof(HeaderItem) * GetHeaderItemNumber(db, srcHeaderIndex);
	
	srcH = DmQueryRecord(db, srcHeaderIndex);
	if (srcH == NULL) {
		ErrAlert(DmGetLastErr());
		return;
	}
	srcP = MemHandleLock(srcH);
	
	destH = DmResizeRecord(db, destHeaderIndex, newSize);
	if (destH == NULL) {
		ErrAlert(DmGetLastErr());
		return;
	}
	
	destP = MemHandleLock(destH);
	
	//	Important! dont overwrite header flag bit
	//	Important! plus offset!!!!
	offsetP = &srcP->itemNumbers;
	DmWrite(destP, sizeof(Header), offsetP, newSize - sizeof(Header));
	
	MemPtrUnlock(srcP);	
	MemPtrUnlock(destP);
}

//	search all SGMM database and stores the cardNo&dbID in globals ListH&DBListCount;
//	used in test queue update in appstart() and mainlist in mainform
void GetDBList(){
}

//	check if a word record contains data
Boolean WordRecordContainsData (WordRecordType* recordP)
{
	UInt16 i;

	// Look for a field which isn't empty
	for (i = WordFieldQuestion; i < WordFieldsCount; i++)
	{
		if (recordP->field[i] != NULL)
			return true;
	}

	return false;
}

//	find the proper position of a string as question field, used in browse form
//	for fast jump
UInt16 InstantLookupPosition(DmOpenRef db, Char* p){
	
	WordRecordType record;
	WordPackedRecord* packed;
	MemHandle recordH;
	UInt16 i, newIndex;
	
	//prepare a psuedo-record
	record.header.allBits = BitAtPosition(RawQueue);
	for (i = WordFieldQuestion; i < WordFieldsCount; i ++) record.field[i] = NULL;
	record.field[WordFieldQuestion] = p;
		
	recordH = DmNewHandle(db, WordUnpackedSize(&record));
	
	packed = MemHandleLock(recordH);
	PackRecord(&record, packed);
	
	newIndex = DmFindSortPosition(db, packed, NULL, (DmComparF*)WordDBComparePackedRecords, NULL);
	MemPtrUnlock(packed);
	
	MemHandleFree(recordH);

	if (newIndex > 0) newIndex --;
	return newIndex;
}


void UnpackSuperMemoRecord(DmOpenRef db, UInt16 index, WordRecordType *record, MemHandle* recordH, char* cache){
	
	char* recordP;
	char* p;
	
	(*recordH) = DmQueryRecord(db, index);
	
	if (*recordH == NULL) return;
	
	recordP = MemHandleLock(*recordH);
	
	p = recordP + 20;
		
	record->field[WordFieldQuestion] = p;
	p += StrLen(p) + 1;
	
	StrPrintF(cache, "%s", p);	
	TranslatePhonetic(cache);
	record->field[WordFieldPhonetic] = cache;
	p += StrLen(p) + 1;
	
	record->field[WordFieldAnswer] = p;
	p += StrLen(p) + 1;
	
	record->field[WordFieldNote3] = p;
	

}

static void TranslatePhonetic(char* field){

	char* p;
	UInt16 i, length;

	if (field == NULL || *field == '\0') return;

	length = StrLen(field);
	p = field;
	
	for (i = 0; i < length; i++){
		
		switch((*p)){
		
			case '/':
				(*p) = 'e';
				break;
			
			case '0':
				if (*(p + 1) == ':') (*p) = 'O';
				else (*p) = 'D';
				break;
				
			case '1':
			case 'a':
				if (*(p + 1) == ':') (*p) = 'a';
				else (*p) = 'R';
				break;
			
			case '2':
				if (*(p + 1) == ':') (*p) = 'B';
				else (*p) = 'E';
				break;
			
			case '3':
				(*p) = 'A';
				break;
			
			case '4':
				(*p) = 'Q';
				break;
				
			case '5':
				(*p) = 'T';
				break;
				
			case '6':
				(*p) = 'S';
				break;
			
			case '7':
				(*p) = 'Z';
				break;
			
			case '8':
			 	(*p) = 'N';
			 	break;
			 	
			case '9':
				(*p) = 'X';
				break;
			
			case 'i':
				if (*(p + 1) == ':') (*p) = 'i';
				else (*p) = 'I';
				break;
				
			case 'u':
				if (*(p + 1) == ':') (*p) = 'u';
				else (*p) = 'U';
				break;	
		
			default:
				break;
						
		}
		
		p++;
	}
}

void PushHeaderBuffer(DmOpenRef db, HeaderIndex bufferIndex, UInt16 dbIndex){

	UInt16 index;
	
	if (FindHeaderItemByDBIndex(db, bufferIndex, dbIndex, &index)) return;
	
	AddHeaderItemByDBIndex(db, bufferIndex, dbIndex);
	
	if (GetHeaderItemNumber(db, bufferIndex) > HeaderBufferVolumn)
		RemoveHeaderItem(db, bufferIndex, 0);
}

void ValidateHeaderRecord(DmOpenRef db){
	
	HeaderRecord* header;
	MemHandle h;
	UInt16 num;
	UInt16 flag;
	
	//Testing Record 0
	h = DmQueryRecord(db, TestingRecordIndex);
	if (h == NULL){
		ErrAlert(DmGetLastErr());
		return;
	}
	
	header = MemHandleLock(h);
	flag = 0;
	SetBitMacro(flag, HeaderFlag);
	SetBitMacro(flag, TestingQFlag);
	
	if (header->header.allBits != flag)
		CreateHeaderRecord(db, TestingRecordIndex);
		
	MemHandleUnlock(h);

	//Learning Record 1
	h = DmQueryRecord(db, LearningRecordIndex);
	if (h == NULL){
		ErrAlert(DmGetLastErr());
		return;
	}
	
	header = MemHandleLock(h);
	flag = 0;
	SetBitMacro(flag, HeaderFlag);
	SetBitMacro(flag, LearningQFlag);
	
	if (header->header.allBits != flag)
		CreateHeaderRecord(db, LearningRecordIndex);
		
	MemHandleUnlock(h);
	
	//Testing Buffer 2
	h = DmQueryRecord(db, TestingChoiceBuffer);
	if (h == NULL){
		ErrAlert(DmGetLastErr());
		return;
	}
	
	header = MemHandleLock(h);
	flag = 0;
	SetBitMacro(flag, HeaderFlag);
	SetBitMacro(flag, TestingBFlag);
	
	if (header->header.allBits != flag)
		CreateHeaderRecord(db, TestingChoiceBuffer);
		
	MemHandleUnlock(h);

	//Learning Buffer 3
	h = DmQueryRecord(db, LearningChoiceBuffer);
	if (h == NULL){
		ErrAlert(DmGetLastErr());
		return;
	}
	
	header = MemHandleLock(h);
	flag = 0;
	SetBitMacro(flag, HeaderFlag);
	SetBitMacro(flag, LearningBFlag);
	
	if (header->header.allBits != flag)
		CreateHeaderRecord(db, LearningChoiceBuffer);
		
	MemHandleUnlock(h);
	
	num = DmNumRecordsInCategory(db, dmAllCategories);
	
	if (num == FirstWordRecord) return;
	
	while(num > FirstWordRecord){
		h = DmQueryRecord(db, FirstWordRecord);
		if (h == NULL){
			ErrAlert(DmGetLastErr());
			return;
		}
	
		header = MemHandleLock(h);
		
		if (GetBitMacro(header -> header.allBits, HeaderFlag)){
			MemHandleUnlock(h);
			DmRemoveRecord(db, FirstWordRecord);
		}
		else {
			MemHandleUnlock (h);
			break;
		}
		
		num --;
	}
}

void UpdateOldVersionDatabase3(DmOpenRef db){

	WordPackedRecord* packed;
	WordPackedRecord* packedHelper = 0;
	SugarAppInfoType* app;	
	MemHandle recordH;
	
	//version can handled by this routine		
	UInt16 num, i, bits;
	Header waitingFlag;
	Header testingFlag;
	LearningStatus ls = FirstLook;
	DateType today;
	UInt32 days; 
	
	UInt32 progress = 0;
	UInt32 newProgress;
	
	Char title[20]="ɰ汾ݿ";
	
	DateSecondsToDate(TimGetSeconds(), &today);
	days = DateToDays(today);
	
	waitingFlag.allBits = 0;
	SetBitMacro(waitingFlag.allBits, WaitingQueue);
	
	testingFlag.allBits = 0;
	SetBitMacro(testingFlag.allBits, TestingQueue);
	
	num = DmNumRecords(db);	

	app = GetAppInfoPtr(db);
	
	if (app == NULL) return;
	
	if (app -> version > 1) {
		MemPtrUnlock(app);
		return;
	}
	
	
	if (num > 0) ProgressPie(progress, title);
	
	for ( i = 0; i < num; i++){

		newProgress = (UInt32)i * ProgressFull / (num - 1);
		if (newProgress != progress) {
			progress = newProgress;
			ProgressPie(progress, title);
		}

		
		//	query record
		recordH = DmQueryRecord(db, i);
		if (recordH == NULL) {
			ErrAlert(DmGetLastErr());
			continue;
		}
	
		packed = MemHandleLock(recordH);
		
		//	check if it is header record, if so, continue;
		bits = packed -> header.allBits;
		
		if ((GetBitMacro(bits, HeaderFlag) != 0) ||	// 	header queue
			(GetBitMacro(bits, 0) != 0) || 			//	raw queue
			(GetBitMacro(bits, 1) != 0)){			// 	learning queue
			
			MemHandleUnlock(recordH);
			continue;
		}
		
		//	old testing queue, if testing then testing, otherwise, waiting;
		if (GetBitMacro(bits, 2)!= 0){ 	//	testing queue
			
			if (DateToDays(packed -> memoStatus.nextTest) > days){	//should be waiting
				
				DmWrite(packed, (UInt32)&packedHelper -> header, &waitingFlag, sizeof(waitingFlag));
				DmWrite(packed, (UInt32)&packedHelper -> ltStatus.learningStatus, &ls, sizeof(ls));
				
				MemHandleUnlock(recordH);
				continue;
			}
			else {
				DmWrite(packed, (UInt32)&packedHelper -> header, &testingFlag, sizeof(testingFlag));
				DmWrite(packed, (UInt32)&packedHelper -> ltStatus.learningStatus, &ls, sizeof(ls));

				MemHandleUnlock(recordH);
				continue;			
			}
		}

		//	old passing queue and failing queue, change to testing		
		if ((GetBitMacro(bits, 3) != 0) || (GetBitMacro(bits, 4))){
			
			DmWrite(packed, (UInt32)&packedHelper -> header, &testingFlag, sizeof(testingFlag));
			DmWrite(packed, (UInt32)&packedHelper -> ltStatus.learningStatus, &ls, sizeof(ls));

			MemHandleUnlock(recordH);
			continue;					
		}
	}	//end for loop
	
	
}

// Added by hytown
//	rebuild testing queue
void ResetTestingStatus(DmOpenRef db){
	
	UInt16 j, num;
	
	UInt32 progress = 0;
	UInt32 newProgress;
	Char title[dmDBNameLength + 20];
	Char dbName[dmDBNameLength + 1] = "Unknown";
	LocalID localID;
	Err error;
	TestingStatus ts;
		
	
	//loop records
	num = DmNumRecordsInCategory(db, dmAllCategories);
	
	if (num > FirstWordRecord) {
		error = DmOpenDatabaseInfo(db, &localID, NULL, NULL, NULL, NULL);
		
		if (error) ErrAlert(error);
		else {
			
			error = DmDatabaseInfo(0, localID, dbName, NULL, NULL, NULL, NULL, NULL, 
				NULL, NULL, NULL, NULL, NULL);
		
			if (error) ErrAlert(error);
			else StrPrintF(title, "%s %s", "Reseting testing status", dbName);
			
		}
		ProgressPie(progress, title);
	}		
	
	for (j = FirstWordRecord; j < num; j++){
			
		ts.firstRating = NoRating;
		ts.secondRating = NoRating;
		ts.totalFail = 0;
		SetTestingStatus(db, j, ts);
		
		newProgress = (j - 1) * ProgressFull / (num - FirstWordRecord);
		if (newProgress != progress) {
			progress = newProgress;
			ProgressPie(progress, title);
		}
	}
	
}
