#include <PalmOS.h>

#include "SugarMemo.h"
#include "SugarMemoDB.h"
#include "SugarMemoBrowse.h"
#include "SugarMemoLearn.h"
#include "SugarMemoLearn2.h"
#include "SugarMemoEdit.h"
#include "SugarMemoDetail.h"
#include "SugarMemoRsc.h"
#include "SugarMemoTools.h"
#include "SugarMemoPrefs.h"
#include "SugarMemoFont.h"

#define RowsInDatabaseTable (Int16)11
#define	IntColumnWidth		(UInt16)32

#define DatabaseNameColumn	0

// Define the minimum OS version we support
#define ourMinVersion    sysMakeROMVersion(3,0,0,sysROMStageDevelopment,0)
#define kPalmOS10Version sysMakeROMVersion(1,0,0,sysROMStageRelease,0)

// ********************************************************************
// Global variables
// ********************************************************************

// g_prefs	cache for application preferences during program execution
SugarMemoPreferenceType g_prefs;

UInt32	DatabaseID[RowsInDatabaseTable];
UInt16 	HeaderRecordIndex = HeaderIndexCounts;

//custom font handle
MemHandle phoneticSmallFontH, phoneticLargeFontH, phoneticHugeFontH, OS5_16_FontH;
FontType* phoneticSmallFontP, *phoneticLargeFontP, *phoneticHugeFontP, *OS5_16_FontP;

// ********************************************************************
// Internal Constants
// ********************************************************************

//*******************************************************************
//	Internal functions
//*******************************************************************

//extern Boolean Learn2DoHardKey(EventType* event);

static void DrawDatabaseName(void* table, Int16 row, Int16 column, RectangleType* bounds);
static void DrawCell(Int32 count, RectangleType* cellBounds);
static void MainFormLoadTable();
static void MainFormInit();

static Boolean GetSelectedDatabaseID(LocalID* id);
static Boolean GetSelectedDatabase();

static void DoCommandNewDatabase();
static void DoCommandRenameDatabase();
static void DoCommandDeleteDatabase();
static void DoCommandResetDatabase();
static void DoCommandImport();
static void ClearChoiceBuffer();
static void SortStudyList();
static void RebuildTestQueue();
static void InitializeAllDatabase();
static void PeekDBVersion();


//Main Form Init
void MainFormInit()
{
	Int16 row;
	TableType* table = GetObjectPtr(MainDatabaseTable);

	for (row = 0; row < RowsInDatabaseTable; row++)
	{
		TblSetItemStyle(table, row, DatabaseNameColumn, customTableItem);
		TblSetRowUsable(table, row, false);
	}

	TblSetColumnUsable(table, DatabaseNameColumn, true);

	// Set the callback routine that will draw the records.
	TblSetCustomDrawProcedure (table, DatabaseNameColumn, DrawDatabaseName);	
}

void MainFormLoadTable(){
	
	TableType* table = GetObjectPtr(MainDatabaseTable);
	Int16 row = 0;
	
	Boolean newSearch = true;
	DmSearchStateType searchState;
	UInt16 cardNo;
	LocalID dbID;
	
	DmOpenRef db;
	
	TblUnhighlightSelection(table);
				
	//	search & retrieve database info, set Column value
	while ((0 == DmGetNextDatabaseByTypeCreator(newSearch, &searchState, appVocabularyDB, 
		appFileCreator, false, &cardNo, &dbID)) && (row < RowsInDatabaseTable)){
	
		newSearch = false;
		
		DatabaseID[row] = dbID;			//	save db ID
		db = DmOpenDatabase(cardNo, dbID, dmModeReadOnly);
		
		if (db == 0) {
			ErrAlert(DmGetLastErr());
		}
		else {
			TblSetRowUsable(table, row, true);
			DmCloseDatabase(db);
		} 
		
		row++;				
	}
	
	//	mark left row unusable
	while (row < RowsInDatabaseTable){
		TblSetRowUsable(table, row, false);
		row ++;
	}

	TblDrawTable(table);	
}

void DrawDatabaseName(void* table, Int16 row, Int16 column, RectangleType* bounds){
	
	Char name[dmDBNameLength + 1];
	DmOpenRef	db;
	FontID	currFont;
	Int32	count;
	RectangleType cellBounds;
	Err error;
	
	error = DmDatabaseInfo(0, DatabaseID[row], name, NULL, NULL, NULL, NULL, NULL,
		NULL, NULL, NULL, NULL, NULL);
		
	if (error) {
		ErrAlert(error);
		StrPrintF(name, "%s", "Error in DmDatabaseInfo()");
	}	
	
	currFont = FntSetFont(boldFont);
	WinDrawChars(name, StrLen(name), bounds ->topLeft.x, bounds ->topLeft.y);

	db = DmOpenDatabase(0, DatabaseID[row], dmModeReadWrite);
	
	if (db == 0) ErrAlert(DmGetLastErr());
	else {
		count = (Int32)GetHeaderItemNumber(db, LearningRecordIndex);
		
		cellBounds.topLeft.x = bounds -> topLeft.x + bounds -> extent.x - 3 * IntColumnWidth;
		cellBounds.extent.x = IntColumnWidth;
		cellBounds.topLeft.y = bounds -> topLeft.y;
		cellBounds.extent.y = bounds -> extent.y;
		
		DrawCell(count, &cellBounds);
		
		count = (Int32)GetHeaderItemNumber(db, TestingRecordIndex);
		cellBounds.topLeft.x += IntColumnWidth;
		DrawCell(count, &cellBounds);
		
		count = (Int32)DmNumRecordsInCategory(db, dmAllCategories) - FirstWordRecord;
		cellBounds.topLeft.x += IntColumnWidth;
		DrawCell(count, &cellBounds);
		
		DmCloseDatabase(db);
	}
}

void DrawCell(Int32 count, RectangleType* cellBounds){
	
	Int16 strWidth;
	Int16 strLength;
	Boolean fitInWidth;
	
	Char num[maxStrIToALen + 1];
	StrIToA(num, count);
	
	strLength = StrLen(num);
	strWidth = cellBounds -> extent.x;
	FntCharsInWidth(num, &strWidth, &strLength, &fitInWidth);
	
	WinDrawChars(num, strLength, cellBounds -> topLeft.x + (cellBounds -> extent.x - strWidth)/2, cellBounds -> topLeft.y);
}

Boolean GetSelectedDatabaseID(LocalID* id){

	TableType* table = GetObjectPtr(MainDatabaseTable);
	Int16 row, column;
	
	if (!TblGetSelection(table, &row, &column)){
	
		DoCustomDialog("Sorry", "Please select a database from the list first.", 
			false, 0, NULL, false);
		return false;
	
	}

	*id = DatabaseID[row];
	return true;
}


Boolean GetSelectedDatabase(){

	LocalID localID;
	
	if (!GetSelectedDatabaseID(&localID)) return false;
	
	CurrentDB = DmOpenDatabase(0, localID, dmModeReadWrite);
	
	if (CurrentDB == 0){
		ErrAlert(DmGetLastErr());
		return false;
	}
		
	return true;
}

void DoCommandNewDatabase(){
	
	Char dbName[dmDBNameLength + 1];
	
	MenuEraseStatus(0);
	
	if (DoCustomDialog("New Database", "Input new database name", true, dmDBNameLength, dbName, true)){
		CreateWordDB(dbName);	
		MainFormLoadTable();
    }
}

void DoCommandRenameDatabase(){
	
	LocalID localID;
	Char dbName[dmDBNameLength + 1];
	Err error;
	
	MenuEraseStatus(0);
	
	if (!GetSelectedDatabaseID(&localID)) return;
			
	if (DoCustomDialog("Rename database", "Input new name", true, dmDBNameLength, dbName, true)){
		
		error = DmSetDatabaseInfo(0, localID, dbName, NULL, NULL, 
    				NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    	
    	if (error) ErrAlert(error);
    	else MainFormLoadTable();
		
	}
}

void DoCommandDeleteDatabase(){
	
	TableType* table = GetObjectPtr(MainDatabaseTable);
	LocalID localID;
	Char dbName[dmDBNameLength + 1];
	Err error;
	
	MenuEraseStatus(0);
	
	if (!GetSelectedDatabaseID(&localID)) return;
	
	error = DmDatabaseInfo(0, localID, dbName, NULL, NULL, NULL, NULL, NULL, 
		NULL, NULL, NULL, NULL, NULL);
		
	if (error){
		ErrAlert(error);
		return;
	}

	if (DoCustomDialog(dbName, "Are you sure to delete the selected database?", false, 0, NULL, true)){
		TblUnhighlightSelection(table);
		DeleteWordDB(0, localID);	
		MainFormLoadTable();
	}
}

void DoCommandResetDatabase(){

	LocalID localID;
	DmOpenRef db;

	if (!GetSelectedDatabaseID(&localID)) return;
	
	if (DoCustomDialog("Reset Database", "This operation will clear the memorizing status of all words, continue?", 
		false, 0, NULL, true)){
    	
    	db = DmOpenDatabase(0, localID, dmModeReadWrite);
    	
    	if (db == 0) {   		
    		ErrAlert(DmGetLastErr());
    		return;
    	}
    	
    	ResetDatabase(db);
   		DmCloseDatabase(db);
   		
		MainFormLoadTable();
	}
}

void ClearChoiceBuffer(){

	LocalID localID;
	DmOpenRef db;
	
	if (!GetSelectedDatabaseID(&localID)) return;

	if (DoCustomDialog("Confirm", "This operation wont clear any memorizing information. Use this operation if you encounter invalide record unique ID error",
		false, NULL, NULL, true)){
		
		db = DmOpenDatabase(0, localID, dmModeReadWrite);
		
		if (db == 0) {
			ErrAlert(DmGetLastErr());
			return;
		}
		
		ClearHeaderRecord(db, LearningChoiceBuffer);	
		ClearHeaderRecord(db, TestingChoiceBuffer);
		DmCloseDatabase(db);
		
		MainFormLoadTable();
	}

}

void SortStudyList(){

	LocalID localID;
	DmOpenRef db;
	
	if (!GetSelectedDatabaseID(&localID)) return;

	if (DoCustomDialog("Confirm", "This operation will sort and rebuild study list.\n\nWARNING: This will clear the learning status of all words in study list.",
		false, NULL, NULL, true)){
		
		db = DmOpenDatabase(0, localID, dmModeReadWrite);
		
		if (db == 0) {
			ErrAlert(DmGetLastErr());
			return;
		}
		
		RebuildLearningQueue(db);	
		DmCloseDatabase(db);
		
		MainFormLoadTable();
	}
}

void RebuildTestQueue(){

	LocalID localID;
	DmOpenRef db;
	
	if (!GetSelectedDatabaseID(&localID)) return;

	if (DoCustomDialog("Confirm", "This operation will rebuild and repair testing list.",
		false, NULL, NULL, true)){

		db = DmOpenDatabase(0, localID, dmModeReadWrite);
		
		if (db == 0) {
			ErrAlert(DmGetLastErr());
			return;
		}
		
		RebuildTestingQueue(db, true);
		DmCloseDatabase(db);
		
		MainFormLoadTable();
	}
}

void InitializeAllDatabase(){

	SugarAppInfoType* app, *appHelper = 0;
	TableType* table = GetObjectPtr(MainDatabaseTable);
	Int16 i, lastRow;
	UInt16 version, bits;
	DmOpenRef db;
	Err error;
	
	lastRow = TblGetLastUsableRow(table);
	
	if (lastRow == tblUnusableRow) return;
	
	for (i = 0; i <= lastRow; i++){
	
		db = DmOpenDatabase(0, DatabaseID[i], dmModeReadWrite);
		
		if (db == 0) {
			ErrAlert(DmGetLastErr());
			continue;
		}
		
		UpdateOldVersionDatabase3(db);
		//DoCustomDialog("Debug", "Old Version Database updated", false, 0, NULL, false);
		

		ValidateHeaderRecord(db);
		//DoCustomDialog("Debug", "HeaderRecord validated", false, 0, NULL, false);

		error = SugarDBAppInfoInit(db);		
		if (error){
			ErrAlert(error);			
			DmCloseDatabase(db);
			continue;
		}
		
		bits = 0;
		SetBitMacro(bits, WordFieldPhonetic);
		SetBitMacro(bits, WordFieldNote3);

		app = GetAppInfoPtr(db);
		if (app -> answerPageFields.allBits != bits){
		
			DmWrite(app, (UInt32)&appHelper -> answerPageFields.allBits, &bits, sizeof(bits));
		
		}
		MemPtrUnlock(app);
		
		RebuildTestingQueue(db, false);
		
		// Added by hytown
//		ResetTestingStatus(db);
		
		//DoCustomDialog("Debug", "successfully rebuild testing queue", false, 0, NULL, false);
		DmCloseDatabase(db);
	}

	// Added by hytown to update database infomation	
	if (lastRow >= 0) TblSelectItem(table, 0, 0); // Seems not working
	TblDrawTable(table);

}

void PeekDBVersion(){
	
	SugarAppInfoType* app;
	LocalID localID;
	DmOpenRef db;
	Char buffer[maxStrIToALen + 20];
	
	if (!GetSelectedDatabaseID(&localID)) return;

	db = DmOpenDatabase(0, localID, dmModeReadWrite);
		
	if (db == 0) {
		ErrAlert(DmGetLastErr());
		return;
	}
	
	app = GetAppInfoPtr(db);
	
	if (app != NULL) {
		StrPrintF(buffer, "%s %d", "DB Version:", app -> version);
		MemPtrUnlock(app);
		
		ProgressPie(0, buffer);
		SysTaskDelay(60);
		ProgressPie(ProgressFull, buffer);
	}
		
	DmCloseDatabase(db);		
}

void DoCommandImport(){

	Char name[dmDBNameLength + 1];
	Char newName[dmDBNameLength + 1];

	Char cache[200];
	
	DmOpenRef db, newdb;
	LocalID localID;
	UInt16 num, index, newIndex;
	UInt16 attr = dmUnfiledCategory;
	
	WordRecordType defaultNewRecord, record;
	MemHandle recordH;

	Boolean result;
	

	result = DoCustomDialog("תݿ", "ݿ", 
		true, dmDBNameLength, name, true);

	if (result == false) return;
	
	localID = DmFindDatabase(0, name);
	
	if (localID == 0) {
		DoCustomDialog("!@#$%^&*", "Ҳݿ⡣", false, 0, NULL, false);
		return;
	}

    db = DmOpenDatabase(0, localID, dmModeReadWrite);
    			
    num = DmNumRecordsInCategory(db, dmAllCategories);
    			
    if (num < 17) {
    	DmCloseDatabase(db);
    	return;		
    }
    		
    StrPrintF(newName, "%s%s", "SG_", name);
    CreateWordDB(newName);
    		
    localID = DmFindDatabase(0, newName);
    newdb = DmOpenDatabase(0, localID, dmModeReadWrite);
    		
	MemSet(&defaultNewRecord, sizeof(Header) + sizeof(LTStatus) + sizeof(MemorizingStatus), 0);	
	SetBitMacro(defaultNewRecord.header.allBits, RawQueue);
	defaultNewRecord.memoStatus.eFactor = 2.5;
	defaultNewRecord.memoStatus.difficulty = 40;
	
	for (index = WordFieldQuestion; index < WordFieldsCount; index++)
		defaultNewRecord.field[index] = NULL;

	for (index = 16; index < num; index++){
				
		record = defaultNewRecord;
				
		UnpackSuperMemoRecord(db, index, &record, &recordH, cache);
				
		NewWordRecord(newdb, &record, &newIndex);

		DmSetRecordInfo (newdb, newIndex, &attr, NULL);
				
		MemHandleUnlock(recordH);
			
	}
			
	DmCloseDatabase(newdb);
	
	MainFormLoadTable();
}

	
/*	
	FormType * form;
	FieldType * field;
	Char* p;
	Char newp[50];
	Char cache[100];
	
	DmOpenRef db, newdb;
	LocalID localID;
	UInt16 num, index, newIndex;
	UInt16 attr = dmUnfiledCategory;
	
	WordRecordType record;
	MemHandle recordH;
			
	form = FrmInitForm(InputDBNameForm);
	field = FrmGetObjectPtr(form, FrmGetObjectIndex(form, InputDBNameInputField));
	FldSetMaxChars(field, DBNameMaxLen);
	
    if (FrmDoDialog(form) == InputDBNameOKButton) {
 
	   	p = FldGetTextPtr(field);

    	if (p != NULL && StrLen(p)>0) {
    		
    		localID = DmFindDatabase(0, p);
    		
    		if (localID == 0) goto end;
    			
    		db = DmOpenDatabase(0, localID, dmModeReadWrite);
    			
    		num = DmNumRecordsInCategory(db, dmAllCategories);
    			
    		if (num < 17) {
    			DmCloseDatabase(db);
    			goto end; 			
    		}
    		
    		StrPrintF(newp, "%s%s", "new_", p);
    		CreateWordDB(newp);
    		
    		localID = DmFindDatabase(0, newp);
    		newdb = DmOpenDatabase(0, localID, dmModeReadWrite);
    		
			for (index = 16; index < num; index++){
				
				record = DefaultNewWordRecord;
				
				UnpackSuperMemoRecord(db, index, &record, &recordH, cache);
				
				NewWordRecord(newdb, &record, &newIndex);

				DmSetRecordInfo (newdb, newIndex, &attr, NULL);
				
				MemHandleUnlock(recordH);
			
			}
			
			DmCloseDatabase(newdb);
    	}
    }
    		
end:FrmDeleteForm(form);
    FrmUpdateForm(MainForm, frmRedrawUpdateCode);
*/


static Boolean MainFormDoCommand(UInt16 command)
{
    Boolean handled = false;
    FormType * frmP;

    switch (command)
    {
    case MainDatabaseSugarMemo:

        // Clear the menu status from the display
        MenuEraseStatus(0);

        // Display the About Box.
        frmP = FrmInitForm (AboutForm);
        FrmDoDialog (frmP);                    
        FrmDeleteForm (frmP);

        handled = true;
        break;
        
    
    case MainDatabaseMenuItem1200:	//½
		DoCommandNewDatabase();
		handled = true;    		//
    	break;
    	
    case MainDatabaseMenuItem1201:	//
    	DoCommandRenameDatabase();
    	handled = true;
    	break;
    
    case MainDatabaseMenuItem1202:	//ɾ
    	DoCommandDeleteDatabase();
    	handled = true;
    	break;
    	
    case MainDatabaseMenuItem1204: 	//Pref
    	MenuEraseStatus(0);	

    	if (GetSelectedDatabase()) FrmGotoForm(PrefsForm);
    	handled = true;
    	break;

	case MainAdvancedMenuItem1000:	//clear choice buffer
		MenuEraseStatus(0);
		
		ClearChoiceBuffer();
		handled = true;
		break;

	case MainAdvancedMenuItem1001:	//sort study list
		MenuEraseStatus(0);
		
		SortStudyList();
		handled = true;
		break;    

	case MainAdvancedMenuItem1002:	//rebuild test queue
		MenuEraseStatus(0);
		
		RebuildTestQueue();
		handled = true;
		break;

    case MainAdvancedMenuItem1004:  //Reset

		DoCommandResetDatabase();
    	handled = true;
    	break;
    }
    
    return handled;
}

//MainForm Handle Event
static Boolean MainFormHandleEvent(EventType * eventP)
{
    Boolean handled = false;
    FormType * frmP = FrmGetActiveForm();
	 
    switch (eventP->eType) 
        {
        case menuEvent:
            return MainFormDoCommand(eventP->data.menu.itemID);

        case frmOpenEvent:
            MainFormInit();
            FrmDrawForm(frmP);
            MainFormLoadTable();
            InitializeAllDatabase();
            handled = true;
            break;
            				            
        case ctlSelectEvent:
    		switch (eventP->data.ctlSelect.controlID){//switch (event->data.ctlSelect.controlID)
    			case MainReviewButton:
    				
    				if (GetSelectedDatabase()){
    					HeaderRecordIndex = TestingRecordIndex;
    					FrmGotoForm(Learn2Form);
    				}
    				handled = true;
					break;
				
				case MainBrowseButton:
					if (GetSelectedDatabase()){
						HeaderRecordIndex = HeaderIndexCounts;
						FrmGotoForm(BrowseForm);
					}
					handled = true;
					break;
				
				case MainLearnButton:
					if (GetSelectedDatabase()){
						HeaderRecordIndex = LearningRecordIndex;
						FrmGotoForm(Learn2Form);
					}
					handled = true;
					break;
				
				case MainHelpButton:
					FrmHelp(MainHelpString);
					handled = true;
					break;
				
				default:
					break;
			}
			break;
			
		case keyDownEvent:
			if (eventP->data.keyDown.chr == chrSmall_I){
				PeekDBVersion();
				handled = true;
			}
			
			if (eventP->data.keyDown.chr == chrPercentSign){
				DoCommandImport();
				handled = true;
			}
			break;

        default:
            break;
        }    
    
    return handled;
}

// FUNCTION: AppHandleEvent
static Boolean AppHandleEvent(EventType * eventP)
{
    UInt16 formId;
    FormType * frmP;
    Boolean handled = false;

    if (eventP->eType == frmLoadEvent)
    {
        // Load the form resource.
        formId = eventP->data.frmLoad.formID;
        frmP = FrmInitForm(formId);
        FrmSetActiveForm(frmP);

        // Set the event handler for the form.  The handler of the 
        // currently active form is called by FrmHandleEvent each 
        // time is receives an event.
        switch (formId)
        {
        case MainForm:
            	FrmSetEventHandler(frmP, MainFormHandleEvent);
            	break;
                    
        case BrowseForm:
        		FrmSetEventHandler(frmP, BrowseFormHandleEvent);
        		break;
        	        		
        case LearnForm:
        		FrmSetEventHandler(frmP, LearnFormHandleEvent);
        		break;
                
        case EditForm:
        		FrmSetEventHandler(frmP, EditFormHandleEvent);
        		break;
        
        case DetailForm:
        		FrmSetEventHandler(frmP, DetailFormHandleEvent);
        		break;
        
        case PhoneticForm:
        		FrmSetEventHandler(frmP, PhoneticFormHandleEvent);
        		break;
        		
        case PrefsForm:
        		FrmSetEventHandler(frmP, PrefsFormHandleEvent);
        		break;
        		
        case Learn2Form:
        		FrmSetEventHandler(frmP, Learn2FormHandleEvent);
        		break;
        		
        case FontForm:
        		FrmSetEventHandler(frmP, FontFormHandleEvent);
        		break;

        default:
            break;

        }
        return true;
    }
    
    return false;
}

// FUNCTION: AppEventLoop
//
// DESCRIPTION: This routine is the event loop for the application.

static void AppEventLoop(void)
{
    UInt16 error;
    EventType event;
    Boolean handled = false;

    do {
        // change timeout if you need periodic nilEvents
        EvtGetEvent(&event, 50); //evtWaitForever);
        
        handled = false;

        if (event.eType == keyDownEvent){
        		
        	if ((HeaderRecordIndex == LearningRecordIndex) || (HeaderRecordIndex == TestingRecordIndex)){
        	
        		if (TxtCharIsHardKey(event.data.keyDown.modifiers, event.data.keyDown.chr)){
        		
					handled = Learn2FormDoHardKey(&event);
					
				}
				
			}
        
        }
       
        if (!handled)
        {
        	if (! SysHandleEvent(&event))
        	{

            	if (! MenuHandleEvent(0, &event, &error))
            	{
                	if (! AppHandleEvent(&event))
                	{
                    	FrmDispatchEvent(&event);
                	}
            	}
        	}
        }
    } while (event.eType != appStopEvent);
}

// FUNCTION: AppStart
//
// DESCRIPTION:  Get the current application's preferences.
//
// RETURNED:
//     errNone - if nothing went wrong

static Err AppStart(void)
{
    UInt16 prefsSize;

    // Read the saved preferences / saved-state information.
    prefsSize = sizeof(SugarMemoPreferenceType);
    if (PrefGetAppPreferences(
        appFileCreator, appPrefID, &g_prefs, &prefsSize, true) != 
        noPreferenceFound)
    {
        // FIXME: setup g_prefs with default values
    }
        
    phoneticSmallFontH = DmGetResource('NFNT', 12345);
    phoneticSmallFontP = (FontType*) MemHandleLock(phoneticSmallFontH);
    FntDefineFont(PhoneticSmallFont, phoneticSmallFontP);

    phoneticLargeFontH = DmGetResource('NFNT', 12346);
    phoneticLargeFontP = (FontType*) MemHandleLock(phoneticLargeFontH);
    FntDefineFont(PhoneticLargeFont, phoneticLargeFontP);

    phoneticHugeFontH = DmGetResource('NFNT', 12347);
    phoneticHugeFontP = (FontType*) MemHandleLock(phoneticHugeFontH);
    FntDefineFont(PhoneticHugeFont, phoneticHugeFontP);
    
    OS5_16_FontH = DmGetResource('NFNT', 12348);
    OS5_16_FontP = (FontType*) MemHandleLock(OS5_16_FontH);
    FntDefineFont(OS5_16_Font, OS5_16_FontP);
    	
    return errNone;
}

// FUNCTION: AppStop
static void AppStop(void)
{
	
    // Write the saved preferences / saved-state information.  This 
    // data will be saved during a HotSync backup.
    PrefSetAppPreferences(
        appFileCreator, appPrefID, appPrefVersionNum, 
        &g_prefs, sizeof(SugarMemoPreferenceType), true);
    
    MemHandleUnlock(phoneticSmallFontH);
    DmReleaseResource(phoneticSmallFontH);    

    MemHandleUnlock(phoneticLargeFontH);
    DmReleaseResource(phoneticLargeFontH);    
    
    MemHandleUnlock(phoneticHugeFontH);
    DmReleaseResource(phoneticHugeFontH);
    
    MemHandleUnlock(OS5_16_FontH);
    DmReleaseResource(OS5_16_FontH);

    // Close all the open forms.
    FrmCloseAllForms();    
}

// all code from here to end of file should use no global variables
#pragma warn_a5_access on

// FUNCTION: RomVersionCompatible
static Err RomVersionCompatible(UInt32 requiredVersion, UInt16 launchFlags)
{
    UInt32 romVersion;

    // See if we're on in minimum required version of the ROM or later.
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    if (romVersion < requiredVersion)
    {
        if ((launchFlags & 
            (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
            (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
        {
            FrmAlert (RomIncompatibleAlert);

            // Palm OS 1.0 will continuously relaunch this app unless 
            // we switch to another safe one.
            if (romVersion <= kPalmOS10Version)
            {
                AppLaunchWithCommand(
                    sysFileCDefaultApp, 
                    sysAppLaunchCmdNormalLaunch, NULL);
            }
        }

        return sysErrRomIncompatible;
    }

    return errNone;
}

// FUNCTION: SugarMemoPalmMain
static UInt32 SugarMemoPalmMain(
    UInt16 cmd, 
    MemPtr /*cmdPBP*/, 
    UInt16 launchFlags)
{
    Err error;

    error = RomVersionCompatible (ourMinVersion, launchFlags);
    if (error) return (error);

    switch (cmd)
    {
    case sysAppLaunchCmdNormalLaunch:
        error = AppStart();
        if (error) 
            return error;

        // start application by opening the main form
        // and then entering the main event loop
        FrmGotoForm(MainForm);
        AppEventLoop();
        
        AppStop();
        break;

    default:
        break;
    }

    return errNone;
}

// FUNCTION: PilotMain
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
    return SugarMemoPalmMain(cmd, cmdPBP, launchFlags);
}

// turn a5 warning off to prevent it being set off by C++
// static initializer code generation
#pragma warn_a5_access reset